// ////////////////////////////////////////////////////
//
// StdView.cpp : implementation of the CStdView class
// Based on class COGLView : public CView
// Inspect COGLView for standard OpenGL initialization
//
////////////////////////////////////////////////////////
#include "stdafx.h"
#include "Std.h"

#include "StdDoc.h"
#include "StdView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////
// You can put definitions, variables and helperfunctions
// here. Functions that does not rely on CStdView member
// variables or functions

//////////////////////////////////////////
// cop dimensions
#define BASIC_R		10.0f			// radius platform
#define BASIC_L		1.0f			// length of platform (thickness)
#define BODY_R		BASIC_R/3.0f	// radius of body
#define BODY_L		15.0f			// length of body
#define ARM_R		0.5f			// radius of all armparts
#define ARM_L		BODY_L/3.0f		// length of all armparts
#define HAND_R		3.0f			// hand radius
#define HAND_L		0.5f			// hand length (thickness)
#define NECK_R		ARM_R			// radius of neck		
#define NECK_L		2.0f			// length of neck
#define HEAD_R		HAND_R			// radius of head
#define JOINT_R		ARM_R*1.5f		// radius of joining spheres	


// the lamp above the robocop
#define LAMP_R		HAND_R			// lower radius of conic lamp
#define LAMP_H		HAND_R			// height of lamp

#define QP			20				// quadric precision

// two positions for the movement
static GLfloat v1[6]={0.0f,0.0f,-60.0f,0.0f,60.0f,0.0f};
static GLfloat v2[6]={180.0f,0.0f,60.0f,0.0f,60.0f,0.0f};

// a timer ID
#define MYTIMER 3

// hitname for spotlight
#define LAMP_NAME	17

// hitname for anything else and nothing
#define NO_OBJECT	99

// size of buffer for hitreport
#define BUFSIZE		128


////////////////////////////////////////////////////////////////////////////
void DrawClosedCylinder(GLfloat R,GLfloat H)
{
	// draws a closed cylinder with length along z
	// and leaves origo on top of cylinder
	GLUquadricObj *pQ;
	pQ = gluNewQuadric();

	gluDisk(pQ,0.0f,R,QP,QP);
	gluCylinder(pQ,R,R,H,QP,QP);
	glTranslatef(0.0f,0.0f,H);
	gluDisk(pQ,0.0f,R,QP,QP);

	gluDeleteQuadric(pQ);
}

void DrawRoboCop(GLfloat v[], GLint dim)
{
	// make sure we are within limits
	ASSERT(dim==6);
	
	glPushMatrix();// make this drawing independant

	// need a quadric for cop-parts
	GLUquadricObj *pQ;
	pQ = gluNewQuadric();

	// cop is placed on xy-plane
	// head up twds z

	// draw basic disk
	CMaterial::SetMaterial(EMERALD);
	DrawClosedCylinder(BASIC_R,BASIC_L);
	// rotate body
	glRotatef(v[0],0.0f,0.0f,1.0f);
	// draw body
	DrawClosedCylinder(BODY_R,BODY_L);
	// remember the top of body
	glPushMatrix();
		// Rotate neck and head
		glRotatef(v[5],0.0f,0.0f,1.0f);
		// draw neck
		DrawClosedCylinder(NECK_R,NECK_L);
		// draw head
		glTranslatef(0.0f,0.0f,HEAD_R);
		gluSphere(pQ,HEAD_R,QP,QP);
		// Get back to top of body
	glPopMatrix();
	// remember the top of body
	glPushMatrix();
		// draw arm 1
		// position shoulder
		glTranslatef(0.0f,BODY_R,-ARM_R);
		// draw joint
		gluSphere(pQ,JOINT_R,QP,QP);
		// rotate shoulder
		glRotatef(-90.0f-v[1],1.0f,0.0f,0.0f);
		// draw armepart upper
		DrawClosedCylinder(ARM_R,ARM_L);
		// rotate elbow
		glRotatef(-v[2],1.0f,0.0f,0.0f);
		// draw joint
		gluSphere(pQ,JOINT_R,QP,QP);
		// draw armepart lower
		DrawClosedCylinder(ARM_R,ARM_L);
		// draw hand
		glTranslatef(0.0f,0.0f,HAND_R);
		glRotatef(90.0f,0.0f,1.0f,0.0f);
		DrawClosedCylinder(HAND_R,HAND_L);
		// Get back to top of body
	glPopMatrix();
	// remember the top of body
	glPushMatrix();
		// draw arm 2
		// position shoulder
		glTranslatef(0.0f,-BODY_R,-ARM_R);
		// draw joint
		gluSphere(pQ,JOINT_R,QP,QP);
		// rotate shoulder
		glRotatef(90.0f+v[3],1.0f,0.0f,0.0f);
		// draw armepart upper
		DrawClosedCylinder(ARM_R,ARM_L);
		// rotate elbow
		glRotatef(v[4],1.0f,0.0f,0.0f);
		// draw joint
		gluSphere(pQ,JOINT_R,QP,QP);
		// draw armepart lower
		DrawClosedCylinder(ARM_R,ARM_L);
		// draw hand
		glTranslatef(0.0f,0.0f,HAND_R);
		glRotatef(90.0f,0.0f,1.0f,0.0f);
		DrawClosedCylinder(HAND_R,HAND_L);
		// Get back to top of body
	glPopMatrix();

	//delete the quadric
	gluDeleteQuadric(pQ);
	// reset matrix to entrystate
	glPopMatrix();
}

void DrawSpotLight(BOOL isOn)
{
	glPushMatrix();
	// Sets up and draws a lamp positioned in
	// origo and spreading its light twds neg.z-axis
	// 60 degrees. 
	// Lamp looks like a cone
	// need a quadric for lamp
	GLUquadricObj *pQ;
	pQ = gluNewQuadric();

	// draw the lamp itself
	glPushMatrix();
	glTranslatef(0.0f,0.0f,LAMP_H);
	gluCylinder(pQ,LAMP_R,0.0f,LAMP_H,QP,QP);
	glPopMatrix(); // get origo back

	//delete the quadric
	gluDeleteQuadric(pQ);

	if(isOn)
	{
		// prepare the light, use GL_LIGHT3

		GLfloat spambient[] =	{0.2f,0.2f,0.2f,1.0f };
		GLfloat spdiffuse[] =	{0.8f,0.8f,0.8f,1.0f };
		GLfloat spspecular[] =	{0.8f,0.8f,0.8f,1.0f };
		GLfloat spposition[] =	{0.0f,0.0f,0.0f,1.0f};
		GLfloat spdirection[] = {0.0f,0.0f,-1.0f};
		GLfloat	spotangle=		60.0f;
		
		glLightfv(GL_LIGHT3, GL_AMBIENT,		spambient);
		glLightfv(GL_LIGHT3, GL_DIFFUSE,		spdiffuse);
		glLightfv(GL_LIGHT3, GL_SPECULAR,		spspecular);
		glLightfv(GL_LIGHT3, GL_POSITION,		spposition);
		glLightfv(GL_LIGHT3, GL_SPOT_DIRECTION,	spdirection);
		glLightf (GL_LIGHT3, GL_SPOT_CUTOFF,	spotangle);

		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT3);
	}
	else
		glDisable(GL_LIGHT3);

	glPopMatrix();
}


//////////////////////////////////////////////////////////
// CStdView

IMPLEMENT_DYNCREATE(CStdView, COGLView)

BEGIN_MESSAGE_MAP(CStdView, COGLView)
	//{{AFX_MSG_MAP(CStdView)
	ON_COMMAND(ID_ACTION_MOVE1, OnActionMove1)
	ON_WM_TIMER()
	ON_COMMAND(ID_ACTION_SPOT, OnActionSpot)
	ON_UPDATE_COMMAND_UI(ID_ACTION_SPOT, OnUpdateActionSpot)
	ON_COMMAND(ID_ACTION_DAY, OnActionDay)
	ON_UPDATE_COMMAND_UI(ID_ACTION_DAY, OnUpdateActionDay)
	ON_WM_RBUTTONDOWN()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

///////////////////////////////////////////////////////////
// CStdView construction/destruction

CStdView::CStdView()
{
	// initialize angles, see DrawScene
	m_xv=0.0f;
	m_yv=0.0f;
	// initial position
	m_v[0]=0.0f;
	m_v[1]=30.0f;
	m_v[2]=30.0f;
	m_v[3]=30.0f;
	m_v[4]=30.0f;
	m_v[5]=30.0f;

	m_SpotIsOn=FALSE;
	m_DayIsOn=FALSE;

	m_antallsteg=10;
	m_t=2000.0f;

	m_hRC=NULL;
	m_hDC=NULL;
};

CStdView::~CStdView()
{}

////////////////////////////////////////////////////////////
// CStdView diagnostics
// You should usually not touch these:
#ifdef _DEBUG
void CStdView::AssertValid() const
{COGLView::AssertValid();}
void CStdView::Dump(CDumpContext& dc) const
{COGLView::Dump(dc);}

CStdDoc* CStdView::GetDocument() // non-debug version is inline
{	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CStdDoc)));
	return (CStdDoc*)m_pDocument;
}
#endif //_DEBUG

//////////////////////////////////////////////////////////////
// CStdView operations

void CStdView::DrawScene()
{
	// Clear the color and depth buffers.
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		
	// Set up for scene
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	gluLookAt(40.0f,0.0f,40.0f,
			0.0f,0.0f,BODY_L,
			0.0f,0.0f,1.0f);

	glLoadName(NO_OBJECT);
	// handle daylight
	if(m_DayIsOn)
		glEnable(GL_LIGHT0);
	else
		glDisable(GL_LIGHT0);

	// rotate around x-axis 
	glRotatef(m_xv,1.0f,0.0f,0.0f);
	// rotate around z-axis 
	glRotatef(m_yv,0.0f,0.0f,1.0f);

	// make ogl nomalize normals
	glEnable(GL_NORMALIZE);

	// draw the lamp
	glLoadName(LAMP_NAME);
	CMaterial::SetMaterial(BRONZE);
	glPushMatrix();
	glTranslatef(0.0f,0.0f,2*BODY_L);
	DrawSpotLight(m_SpotIsOn);
	glPopMatrix();
	
	glLoadName(NO_OBJECT);
	// draw the cop
	DrawRoboCop(m_v,6);
	// finnish drawing
	glFlush();
}

void CStdView::InitializeRContext()
{
	// Here we set the attributes that will be constant
	// during the rendering contexts lifetime
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    
	// prepare ligthsource
	GLfloat ambient[] = {0.2f,0.2f,0.2f,1.0f };
    GLfloat diffuse[] = {0.7f,0.7f,0.7f,1.0f };
    GLfloat position[] = {20.0f,30.0f,7.0f,0.0f };
    GLfloat lmodel_ambient[] = {0.2f,0.2f,0.2f,1.0f };


    glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
    glLightfv(GL_LIGHT0, GL_POSITION, position);
	
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHTING);
	
	// smooth the drawing
	glShadeModel(GL_SMOOTH);

    // set background to deep blue
	glClearColor(0.8f, 0.8f, 0.8f, 0.0f);
}

void CStdView::SetPerspective(int cx,int cy)
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	gluPerspective(60.0f, (GLdouble)cx/cy, 0.1f, 1000.0f);
	//glOrtho(-10.0f,10.0f,-10.0f,10.0f,-10.f,100.0f);
	glMatrixMode(GL_MODELVIEW);
}


void CStdView::OnActionMove1() 
{
	// To make it simple and clean we allways
	// start with setting rob to first position, v1.
	for(int ix=0;ix<6;ix++) m_v[ix]=v1[ix];
	RedrawWindow();

	// There is of course a lot of ways to control
	// number of timer events.I use the fact that 
	// the total time should be 2 s=2000 ms, and 
	// increment a counter,member variable m_t, 
	// by 100 ms for each snapshot,see timer func
	
	// Start with:
	m_acct=0.0f;

	// and go
	SetTimer(MYTIMER,UINT(m_t/m_antallsteg),NULL);
}

void CStdView::OnTimer(UINT nIDEvent) 
{
	// Assume only MYTIMER active
	// Make a move and force redraw
	// if we are not finished
	if(m_acct>m_t)
		KillTimer(MYTIMER);
	else
	{
		m_acct+=m_t/m_antallsteg;
		for(int ix=0;ix<6;ix++)
			m_v[ix]=v1[ix]+(v2[ix]-v1[ix])*(m_acct/m_t);
		RedrawWindow();
	}
}

void CStdView::OnActionSpot() 
{
	// Togles spotlight over robocop
	m_SpotIsOn=!m_SpotIsOn;
	RedrawWindow();
}

void CStdView::OnUpdateActionSpot(CCmdUI* pCmdUI) 
{
	// Shows status of spotlight
	pCmdUI->SetCheck(m_SpotIsOn);
}

void CStdView::OnActionDay() 
{
	// Toggles daylight
	m_DayIsOn=!m_DayIsOn;
	RedrawWindow();
}

void CStdView::OnUpdateActionDay(CCmdUI* pCmdUI) 
{
	// Shows status of daylight (on/of)
	pCmdUI->SetCheck(m_DayIsOn);
}

void CStdView::OnRButtonDown(UINT nFlags, CPoint point) 
{
	// Right button is pressed
	// presumably to toggle spotlight
	GLint n=PickObject(point.x,point.y);
	if (n!=NO_OBJECT)
	{
		// Toggle spotlight
		m_SpotIsOn=!m_SpotIsOn;
		RedrawWindow();
	}
	CView::OnRButtonDown(nFlags, point);
}

int CStdView::PickObject(int x, int y)
{
	// called from right mousedown to identify an object
	// we have only one candidate, the spotlight, so we can
	// consentrate on this

	// set rendering context
	wglMakeCurrent(m_hDC,m_hRC);

	GLuint selectBuf[BUFSIZE];	// to store OGLs reports
	GLint hits;					// to count hits
	GLint viewport[4];			// to store current viewport

	glGetIntegerv(GL_VIEWPORT,viewport);	// get viewport
	glSelectBuffer(BUFSIZE,selectBuf);		// tell OGL where to report
	glRenderMode(GL_SELECT);				// as opposed to GL_RENDER

	glInitNames();
	glPushName(NO_OBJECT); // since -1 gives a warning

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	// set up a small clipping pyramid
	gluPickMatrix(	(GLdouble)x,(GLdouble)(viewport[3]-y),5.0f,5.0f,viewport);
	// same view as we see it, same as in OnSize
	CRect R;
	GetClientRect(&R);
	gluPerspective(	60.0f,(GLdouble)R.Width()/(GLdouble)R.Height(),0.1f,1000.0f);
	glMatrixMode(GL_MODELVIEW);

	DrawScene();// draw without rendering

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	
	hits=glRenderMode(GL_RENDER);// gives me a hitcount
	if(selectBuf[3]==LAMP_NAME)
		return LAMP_NAME;
	else
		return NO_OBJECT;
	wglMakeCurrent(0,0);

}

